/*
 * Copyright 2022 Axel Waggershauser
 */
// SPDX-License-Identifier: Apache-2.0

#pragma once

#include <string>

namespace ZXing {

std::string HRIFromGS1(const std::string& gs1);

} // namespace ZXing
